import os
import torch
import torch.utils.data as data
import torch
import torchvision.transforms as transforms
import random
from PIL import Image, ImageOps
from . import preprocess 
from . import listflowfile as lt
from . import readpfm as rp
import numpy as np
import cv2

IMG_EXTENSIONS = [
    '.jpg', '.JPG', '.jpeg', '.JPEG',
    '.png', '.PNG', '.ppm', '.PPM', '.bmp', '.BMP',
]

def is_image_file(filename):
    return any(filename.endswith(extension) for extension in IMG_EXTENSIONS)

def default_loader(path):
    return Image.open(path).convert('RGB')

def disparity_loader(path):
    return cv2.imread(path, cv2.IMREAD_UNCHANGED)

def disparity_png100_loader(path):
    return cv2.imread(path, cv2.IMREAD_UNCHANGED).astype(np.float32) / 100

class myImageFloder(data.Dataset):
    def __init__(self, left, right, left_disparity, training, th=320, tw=640, aug=False, loader=default_loader, dploader= disparity_loader):
 
        self.left = left
        self.right = right
        self.disp_L = left_disparity
        self.loader = loader
        self.dploader = dploader
        self.training = training
        self.th = th
        self.tw = tw
        print('{} on {}x{}'.format('Train' if training else 'Test', tw, th))
        self.aug = aug

        # create disp loader
        print('Dispatching disp loader for image...')
        self.disp_loaders = []
        for path in self.left:
            if 'DrivingStereo' in path or 'InStereo2K' in path or 'IRSDataset' in path or 'SceneFlow' in path:
                self.disp_loaders.append(disparity_png100_loader)
            else:
                self.disp_loaders.append(disparity_loader)

 
    def __getitem__(self, index):
        left  = self.left[index]
        right = self.right[index]
        disp_L= self.disp_L[index]
        focal = 804.939430

        K = np.array([[focal, 0, 650.577003],
                      [0, focal, 399.905579],
                      [0, 0, 1]])
        K = torch.Tensor(K)

        left_img = self.loader(left)
        right_img = self.loader(right)
        #dataL = self.dploader(disp_L)
        dataL = self.disp_loaders[index](disp_L)
        dataL = np.ascontiguousarray(dataL,dtype=np.float32)

        batch = dict()
        th, tw = self.th, self.tw
        if self.training:  
            w, h = left_img.size

            x1 = random.randint(0, w - tw)
            y1 = random.randint(0, h - th)

            left_img = left_img.crop((x1, y1, x1 + tw, y1 + th))
            right_img = right_img.crop((x1, y1, x1 + tw, y1 + th))

            dataL = dataL[y1:y1 + th, x1:x1 + tw]

            processed = preprocess.get_transform(augment=self.aug)  
            left_img   = processed(left_img)
            right_img  = processed(right_img)

            batch['imgL'], batch['imgR'], batch['disp_true'], batch['left_path'] = left_img, right_img, dataL, left
            batch['K'], batch['x1'], batch['y1'] = K, x1, y1

            return batch
        else:
            processed = preprocess.get_transform(augment=False)
            # resize
            w, h = left_img.size
            scale = float(tw) / w 
            left_img = left_img.resize((tw, th), Image.BILINEAR)
            right_img = right_img.resize((tw, th), Image.BILINEAR)
            dataL = cv2.resize(dataL, (tw, th), interpolation=cv2.INTER_NEAREST) * scale

            left_img       = processed(left_img)
            right_img      = processed(right_img) 
            batch['imgL'], batch['imgR'], batch['disp_true'], batch['left_path'] = left_img, right_img, dataL, left
            batch['K'] = K

            return batch

    def __len__(self):
        return len(self.left)
