import cv2
import sys
import numpy as np

import torch
from torch.utils.data import DataLoader

from ruamel.yaml import YAML

from dataloaders import KITTIRawLoader as KRL

from stereo import Stereo
from onnxsim import simplify
import onnx

torch.backends.cudnn.benchmark = True

torch.set_grad_enabled(False)


def load_configs(path):
    cfg = YAML().load(open(path, 'r'))
    backbone_cfg = YAML().load(
        open(cfg['model']['stereo']['backbone']['cfg_path'], 'r'))
    cfg['model']['stereo']['backbone'].update(backbone_cfg)
    return cfg


if __name__ == '__main__':
    cfg = load_configs(sys.argv[1])

    ckpt = '{}/{}/version_{}/checkpoints/last.ckpt'.format(
        'logs/stereo', cfg['model']['name'], cfg['training']['save_version'])
    cfg['stereo_ckpt'] = ckpt
    print(f'load ckpt from {ckpt}')
    #pose_ssstereo = Stereo.load_from_checkpoint(cfg['stereo_ckpt'],
    #                                            strict=False,
    #                                            cfg=cfg).cuda()

    pose_ssstereo = Stereo(cfg, 'sceneflow').cuda()

    pose_ssstereo.eval()
    dummy = torch.randn(1, 3, 320, 640).cuda()
    output_path = 'CoEx.onnx'
    out = pose_ssstereo(dummy, dummy)
    torch.onnx.export(pose_ssstereo, (dummy, dummy),
                      output_path,
                      verbose=True,
                      export_params=True,
                      do_constant_folding=True,
                      opset_version=11)
    print('simplify model')
    onnx_model = onnx.load(output_path)  # load onnx model
    model_simp, check = simplify(onnx_model)
    assert check, "Simplified ONNX model could not be validated"
    onnx.save(model_simp, output_path + '_sim.onnx')

    exit(1)
