import cv2
import os
import sys
import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from ruamel.yaml import YAML

from dataloaders import KITTIRawLoader as KRL

from stereo import Stereo
from onnxsim import simplify
import onnx

from onnx_tf.backend import prepare
import onnx
import tensorflow as tf

torch.backends.cudnn.benchmark = True

torch.set_grad_enabled(False)


def load_configs(path):
    cfg = YAML().load(open(path, 'r'))
    backbone_cfg = YAML().load(
        open(cfg['model']['stereo']['backbone']['cfg_path'], 'r'))
    cfg['model']['stereo']['backbone'].update(backbone_cfg)
    return cfg


def convert_ReLU_to_LeakyReLU(model, prefix=''):
    for child_name, child in model.named_children():
        #print(prefix + str(type(child)))
        if isinstance(child, nn.LeakyReLU):
            print(prefix + str(type(child)))
            setattr(model, child_name, nn.ReLU())
        else:
            convert_ReLU_to_LeakyReLU(child, prefix + '----')

if __name__ == '__main__':
    cfg_path = sys.argv[1]
    cfg = load_configs(cfg_path)
    
    if len(sys.argv) > 2 and 'ckpt' in sys.argv[2]:
        ckpt = sys.argv[2]
        print(f'Loading ckpt from {ckpt}')
        pose_ssstereo = Stereo.load_from_checkpoint(ckpt, cfg=cfg).cuda()
    else:
        print(f'Dump init param model')
        pose_ssstereo = Stereo(cfg, 'sceneflow')

    if True:
        convert_ReLU_to_LeakyReLU(pose_ssstereo)
    pose_ssstereo.eval()
    dummy = torch.randn(1, 3, 320, 640)
    pose_ssstereo(dummy, dummy)
    output_path = '{}.onnx'.format(os.path.basename(cfg_path)[:-5])

    #exit(1)
    
    torch.onnx.export(pose_ssstereo, (dummy, dummy),
                      output_path,
                      verbose=True,
                      export_params=True,
                      input_names = ['imgL', 'imgR'],
                      output_names = ['disp'],
                      do_constant_folding=True,
                      opset_version=11)
    print('simplify model')
    onnx_model = onnx.load(output_path)  # load onnx model
    model_simp, check = simplify(onnx_model)
    assert check, "Simplified ONNX model could not be validated"
    sim_onnx_path = output_path[:-5] + '_sim.onnx'
    onnx.save(model_simp, sim_onnx_path)
    print('simplify finished')

    TF_PATH = output_path[:-5] + '_tf.pb' # where the representation of tensorflow model will be stored
    onnx_model = onnx.load(sim_onnx_path)  # load onnx model
    tf_rep = prepare(onnx_model, logging_level='DEBUG')  # creating TensorflowRep object
    tf_rep.export_graph(TF_PATH)
    print('export pd finished')
    
    TF_PATH = output_path[:-5] + '_tf.pb' # where the representation of tensorflow model will be stored
    TFLITE_PATH = output_path[:-5] + ".tflite"
    converter = tf.lite.TFLiteConverter.from_saved_model(TF_PATH)
    converter.optimizations = [tf.lite.Optimize.DEFAULT]
    tf_lite_model = converter.convert()
    with open(TFLITE_PATH, 'wb') as f:
        f.write(tf_lite_model)

