import cv2
import os
import sys
import numpy as np

import torch
from torch.utils.data import DataLoader

from ruamel.yaml import YAML

from dataloaders import KITTIRawLoader as KRL

from stereo import Stereo
from onnxsim import simplify
import onnx

from onnx_tf.backend import prepare
import onnx
import tensorflow as tf
from tinynn.converter import TFLiteConverter

torch.backends.cudnn.benchmark = True

torch.set_grad_enabled(False)


def load_configs(path):
    cfg = YAML().load(open(path, 'r'))
    backbone_cfg = YAML().load(
        open(cfg['model']['stereo']['backbone']['cfg_path'], 'r'))
    cfg['model']['stereo']['backbone'].update(backbone_cfg)
    return cfg


if __name__ == '__main__':
    cfg_path = sys.argv[1]
    cfg = load_configs(cfg_path)
    
    if len(sys.argv) > 2 and 'ckpt' in sys.argv[2]:
        ckpt = sys.argv[2]
        print(f'Loading ckpt from {ckpt}')
        model = Stereo.load_from_checkpoint(ckpt, cfg=cfg).cuda()
    else:
        print(f'Dump init param model')
        model = Stereo(cfg, 'sceneflow')

    model.eval()
    dummy = torch.randn(1, 3, 320, 640)
    model(dummy, dummy)

    model_path = 'coex.tflite'

    # Moving the model to cpu and set it to evaluation mode before model conversion
    with torch.no_grad():
        model.cpu()
        model.eval()

        converter = TFLiteConverter(model, (dummy, dummy), model_path)
        converter.convert()
