import cv2
import numpy as np
import glob
import sys
import os

import torch
from torch.utils.data import DataLoader

from ruamel.yaml import YAML

from dataloaders import KITTIRawLoader as KRL

from stereo import Stereo

from dataloaders.stereo import preprocess
from distance_dataset import DistanceDatset

torch.backends.cudnn.benchmark = True


torch.set_grad_enabled(False)

version = 0  # CoEx

vid_date = "2011_09_26"
vid_num = '0093'
half_precision = True


def load_configs(path):
    cfg = YAML().load(open(path, 'r'))
    backbone_cfg = YAML().load(
        open(cfg['model']['stereo']['backbone']['cfg_path'], 'r'))
    cfg['model']['stereo']['backbone'].update(backbone_cfg)
    return cfg


if __name__ == '__main__':
    out = 'result'
    if not os.path.exists(out):
        os.makedirs(out)
    cfg = load_configs(sys.argv[1])

    #ckpt = '{}/{}/version_{}/checkpoints/last.ckpt'.format(
    #    'logs/stereo', cfg['model']['name'], cfg['training']['save_version'])
    ckpt = sys.argv[2]
    cfg['stereo_ckpt'] = ckpt
    print(f'load ckpt from {ckpt}')
    pose_ssstereo = Stereo.load_from_checkpoint(cfg['stereo_ckpt'],
                                                strict=False,
                                                cfg=cfg).cuda()
    pose_ssstereo.eval()
    left_paths = sorted(glob.glob(sys.argv[3]))
    right_paths = sorted(glob.glob(sys.argv[4]))

    fps_list = np.array([])
    processed = preprocess.get_transform(augment=False)
    proc_size = (960, 480)

    dataset = DistanceDatset(dataset='prod_batch1', algorithm='CoEx') # dataset：’beijing’ or ’shenzhen’ or clip_dir
    for leftpath, rightpath in zip(left_paths, right_paths):
        left_img = cv2.imread(leftpath)[:,:,::-1]
        right_img = cv2.imread(rightpath)[:,:,::-1]

        if proc_size is not None:
            raw_size = left_img.shape[:2][::-1]
            disp_scale = float(raw_size[0]) / proc_size[0]
            left_img = cv2.resize(left_img, proc_size)
            right_img = cv2.resize(right_img, proc_size)

        left_input = processed(left_img).unsqueeze(0)
        right_input = processed(right_img).unsqueeze(0)

        start = torch.cuda.Event(enable_timing=True)
        end = torch.cuda.Event(enable_timing=True)
        start.record()

        imgL, imgR = left_input.cuda(), right_input.cuda()
        print(imgL.size())

        end.record()
        torch.cuda.synchronize()
        runtime = start.elapsed_time(end)
        print('Data Preparation: {:.3f}'.format(runtime))

        start = torch.cuda.Event(enable_timing=True)
        end = torch.cuda.Event(enable_timing=True)
        start.record()
        with torch.no_grad():
            with torch.cuda.amp.autocast(enabled=half_precision):
                disp = pose_ssstereo(imgL, imgR, False)
        end.record()
        torch.cuda.synchronize()
        runtime = start.elapsed_time(end)
        # print('Stereo runtime: {:.3f}'.format(runtime))

        fps = 1000/runtime
        fps_list = np.append(fps_list, fps)
        if len(fps_list) > 5:
            fps_list = fps_list[-5:]
        avg_fps = np.mean(fps_list)
        print('Stereo runtime: {:.3f}'.format(1000/avg_fps))


        disp_np = (disp[0]).data.cpu().numpy()
        if proc_size is not None:
            disp_np = cv2.resize(disp_np * disp_scale, raw_size, cv2.INTER_NEAREST)

        clip_name = leftpath.split('/')[-3]
        timestamp = leftpath.split('/')[-1][:-4]
        dataset.collect_by_clip_stamp(disp_np, clip_name, timestamp)

        disp_vis = (disp_np - disp_np.min()) / (disp_np.max() - disp_np.min()) * 255.0
        disp_vis = disp_vis.astype("uint8")
        disp_vis = cv2.applyColorMap(disp_vis, cv2.COLORMAP_INFERNO)
        name = os.path.basename(leftpath)
        save_name = os.path.join(out, name)
         
        #cv2.imwrite(save_name + 'color.jpg', disp_vis)
        #cv2.imwrite(save_name + '.tiff', disp_np)
        
