import cv2
import numpy as np
import glob
import sys
import os
from tqdm import tqdm

import torch
from torch.utils.data import DataLoader
from bag_data import BagData

from ruamel.yaml import YAML

from dataloaders import KITTIRawLoader as KRL

from stereo import Stereo

from dataloaders.stereo import preprocess

torch.backends.cudnn.benchmark = True

torch.set_grad_enabled(False)

half_precision = True

def load_configs(path):
    cfg = YAML().load(open(path, 'r'))
    backbone_cfg = YAML().load(
        open(cfg['model']['stereo']['backbone']['cfg_path'], 'r'))
    cfg['model']['stereo']['backbone'].update(backbone_cfg)
    return cfg


if __name__ == '__main__':
    cfg = load_configs(sys.argv[1])
    bag_list_file = sys.argv[2]
    # pprepare model
    #ckpt = '{}/{}/version_{}/checkpoints/last.ckpt'.format(
    #    'logs/stereo', cfg['model']['name'], cfg['training']['save_version'])
    ckpt=sys.argv[3]
    cfg['stereo_ckpt'] = ckpt
    print(f'load ckpt from {ckpt}')
    pose_ssstereo = Stereo.load_from_checkpoint(cfg['stereo_ckpt'],
                                                strict=False,
                                                cfg=cfg).cuda()
    pose_ssstereo.eval()
    
    # prepare bag
    with open(bag_list_file, 'r') as fin:
        bag_list = [line.strip().split(' ') for line in fin]    

    processed = preprocess.get_transform(augment=False)
    proc_size = (640, 320)
    for bag_file, start_time, label in tqdm(bag_list):
        bag_data = BagData(bag_file, start_time)
        for left_img, right_img, timestamp_ns in bag_data.get_stereo():
            
            if proc_size is not None:
                raw_size = left_img.shape[:2][::-1]
                disp_scale = float(raw_size[0]) / proc_size[0]
                left_img = cv2.resize(left_img, proc_size)
                right_img = cv2.resize(right_img, proc_size)

            left_input = processed(left_img).unsqueeze(0)
            right_input = processed(right_img).unsqueeze(0)

            imgL, imgR = left_input.cuda(), right_input.cuda()
            # print(imgL.size())

            with torch.no_grad():
                with torch.cuda.amp.autocast(enabled=half_precision):
                    disp = pose_ssstereo(imgL, imgR, False)

            disp_np = (disp[0]).data.cpu().numpy()
            # print(disp_np.shape)

            if proc_size is not None:
                disp_np = cv2.resize(disp_np * disp_scale, raw_size, cv2.INTER_NEAREST)
            
            bag_data.write_disparity(disp_np, 50)
        bag_data.release()
            
        
        
